# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: node.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='node.proto',
  package='isek_node',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\nnode.proto\x12\tisek_node\"@\n\x0b\x43\x61llRequest\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x10\n\x08receiver\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\"\x1d\n\x0c\x43\x61llResponse\x12\r\n\x05reply\x18\x01 \x01(\t2J\n\x0fIsekNodeService\x12\x37\n\x04\x63\x61ll\x12\x16.isek_node.CallRequest\x1a\x17.isek_node.CallResponseb\x06proto3'
)




_CALLREQUEST = _descriptor.Descriptor(
  name='CallRequest',
  full_name='isek_node.CallRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sender', full_name='isek_node.CallRequest.sender', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='receiver', full_name='isek_node.CallRequest.receiver', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='isek_node.CallRequest.message', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=89,
)


_CALLRESPONSE = _descriptor.Descriptor(
  name='CallResponse',
  full_name='isek_node.CallResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='reply', full_name='isek_node.CallResponse.reply', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=120,
)

DESCRIPTOR.message_types_by_name['CallRequest'] = _CALLREQUEST
DESCRIPTOR.message_types_by_name['CallResponse'] = _CALLRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CallRequest = _reflection.GeneratedProtocolMessageType('CallRequest', (_message.Message,), {
  'DESCRIPTOR' : _CALLREQUEST,
  '__module__' : 'node_pb2'
  # @@protoc_insertion_point(class_scope:isek_node.CallRequest)
  })
_sym_db.RegisterMessage(CallRequest)

CallResponse = _reflection.GeneratedProtocolMessageType('CallResponse', (_message.Message,), {
  'DESCRIPTOR' : _CALLRESPONSE,
  '__module__' : 'node_pb2'
  # @@protoc_insertion_point(class_scope:isek_node.CallResponse)
  })
_sym_db.RegisterMessage(CallResponse)



_ISEKNODESERVICE = _descriptor.ServiceDescriptor(
  name='IsekNodeService',
  full_name='isek_node.IsekNodeService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=122,
  serialized_end=196,
  methods=[
  _descriptor.MethodDescriptor(
    name='call',
    full_name='isek_node.IsekNodeService.call',
    index=0,
    containing_service=None,
    input_type=_CALLREQUEST,
    output_type=_CALLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_ISEKNODESERVICE)

DESCRIPTOR.services_by_name['IsekNodeService'] = _ISEKNODESERVICE

# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
import click
import importlib.util
import sys
from pathlib import Path

def load_module(script_path: Path):
    """Dynamically load module"""
    try:
       
        module_name = script_path.stem
        spec = importlib.util.spec_from_file_location(module_name, script_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module
    except Exception as e:
        click.secho(f"Module load error: {e}", fg="red")
        sys.exit(1)

def get_available_examples(ctx, args, incomplete):
    """Get available example scripts"""
    examples_dir = Path(__file__).parent.parent / "examples"
    return [
        f.stem for f in examples_dir.glob("*.py") 
        if f.name != "__init__.py" and not f.name.startswith('_')
        and f.stem.startswith(incomplete)
    ]

@click.group()
def cli():
    """ISEK Distributed Multi-Agent Framework CLI"""

@cli.command()
def clean():
    """Run system cleanup script"""
    script_path = Path(__file__).parent.parent / "scripts" / "clean.py"
    module = load_module(script_path)
    module.main()
    click.secho("✓ Cleanup completed", fg="green")

@cli.command()
def registry():
    """Start local development registry"""
    from isek.isek_center import main
    main()

@cli.group()
def example():
    """Example script management"""

@example.command()
def list():
    """List available example scripts"""
    examples_dir = Path(__file__).parent.parent / "examples"
    examples = [
        f.stem for f in examples_dir.glob("*.py") 
        if f.name != "__init__.py" and not f.name.startswith('_')
    ]
    
    click.echo("Available examples:")
    for name in sorted(examples):
        click.echo(f"  • {name}")

@example.command(context_settings=dict(
    ignore_unknown_options=True,
    allow_extra_args=True
))
@click.argument('name', 
    type=click.STRING, 
    shell_complete=get_available_examples
)
@click.pass_context
def run(ctx, name: str):
    """Execute specific example script"""
    script_path = Path(__file__).parent.parent / "examples" / f"{name}.py"
    
    module = load_module(script_path)
    if hasattr(module, 'main'):
        extra_args = ctx.args
        module.main(*extra_args) if extra_args else module.main()
        click.secho(f"✓ Example '{name}' completed", fg="green")
    else:
        click.secho(f"✗ Example '{name}' lacks main() function", fg="red")
        sys.exit(1)

if __name__ == "__main__":
    cli()


from isek.util.logger import LoggerManager, logger

def build(debug=False):
    # 根据 `debug` 重新初始化 Logger
    LoggerManager.init(debug=debug)

    logger.info("Logger initialized with debug mode: {}".format(debug))
    logger.debug("This is a debug message")  # 仅在 debug=True 时显示
    logger.info("This is an info message")


if __name__ == "__main__":
    build(debug=True)


from isek.node.isek_center_registry import IsekCenterRegistry


node_a = IsekCenterRegistry()
node_a.register_node("node_a", "localhost", 8080, {"name": "a"})

node_b = IsekCenterRegistry()
node_b.register_node("node_b", "localhost", 8082, {"name": "b"})

print(node_b.get_available_nodes())
node_a.deregister_node("node_a")
print(node_b.get_available_nodes())



import time
import os
import numpy as np
from isek.agent.distributed_agent import DistributedAgent
from isek.node.etcd_registry import EtcdRegistry
from isek.node.isek_center_registry import IsekCenterRegistry
from isek.agent.persona import Persona
from isek.util.logger import LoggerManager, logger
from isek.llm import OpenAIModel
from dotenv import load_dotenv


def publish_content(platform="twitter",content=""):
    logger.info(f"[]published content on {platform}: {content}")
    return "content published on" + platform
def image_gen(input=""):
    logger.info(f"[]image generated")
    return "image generated with id" + str(np.random.randint(100))
    
def main():
    # registry = EtcdRegistry(host=os.environ["ISEK_ETCD_IP"], port=os.environ["ISEK_ETCD_PORT"])
    
    # run local registry using 
    # python isek_center.py
    # pip install isek
    # how to?
    # isek run registry
    
    #LoggerManager.init(debug=True)
    load_dotenv()
    registry = IsekCenterRegistry()
    Mani_info = {
        "name": "Mani",
        "bio": "An experienced manager",
        "lore": "Your mission is to manage the team",
        "knowledge": "recruit, assign tasks, manage tasks, manage projects",
        "routine": "1. if receive a task, decompose the task to subtasks. 2. search for partners for each task. 3. send message to other agnet to do the tasks. user is your boss, do not ask user to do the task"
    }

    Eddie_info = {
        "name": "Eddie",
        "bio": "An experienced photo designer",
        "lore": "Your mission is to generate beautiful photos based on clients' requirements",
        "knowledge": "photo editing, photo design, photo retouching",
        "routine": ""
    }
    Infy_info = {
        "name": "Infy",
        "bio": "An experienced influencer",
        "lore": "Your mission is to opreate social media accounts and generate traffic",
        "knowledge": "social media, social media marketing, social media management",
        "routine": "you are responsiable for publish content on twitter, facebook, and instagram"
    }

    model = OpenAIModel(
        model_name=os.environ.get("OPENAI_MODEL_NAME"),
        api_key=os.environ.get("OPENAI_API_KEY")
    )

    Mani = Persona.from_json(Mani_info)
    Eddie = Persona.from_json(Eddie_info)
    Infy = Persona.from_json(Infy_info)

    Mani_agent = DistributedAgent(persona=Mani, host="localhost", port=8080, registry=registry, model=model)
    Mani_agent.tool_manager.register_tools([
        Mani_agent.search_partners,
        Mani_agent.send_message,
        Mani_agent.decompose_task,
    ])
    Mani_agent.build(daemon=True)

    Eddie_agent = DistributedAgent(persona=Eddie, host="localhost", port=8081, registry=registry, model=model)
    Eddie_agent.tool_manager.register_tools([
        image_gen,
        Eddie_agent.search_partners,
        Eddie_agent.send_message,
    ])
    Eddie_agent.build(daemon=True)

    Infy_agent = DistributedAgent(persona=Infy, host="localhost", port=8082, registry=registry, model=model)
    Infy_agent.tool_manager.register_tools([
        publish_content,
    ])
    Infy_agent.build(daemon=True)
    
    time.sleep(2)
    # Mani_agent.run("I want to promote my shoes product with a budget of $100")
    
    Mani_agent.run_cli()

    
if __name__ == "__main__":
    main()



import inspect
import re
import json
import hashlib


def function_to_schema(func) -> dict:
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }

    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        try:
            param_type = type_map.get(param.annotation, "string")
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
            )
        parameters[param.name] = {"type": param_type}

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": (func.__doc__ or "").strip(),
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required,
            },
        },
    }


def load_json_from_chat_response(chat_result):
    pattern = r'```json(.*?)```'
    match = re.search(pattern, chat_result, re.DOTALL)
    if match:
        json_data = match.group(1).strip()
        return json.loads(json_data)
    else:
        raise RuntimeError(f"Source str[{chat_result}] no json content extracted.")


def split_list(input_list, chunk_size):
    return [input_list[i:i + chunk_size] for i in range(0, len(input_list), chunk_size)]


def md5(source):
    return hashlib.md5(source.encode()).hexdigest()


def dict_md5(dict_content, exclude_fields=None):
    dict_to_md5 = dict(dict_content)
    exclude_dict = {}
    if exclude_fields and len(exclude_fields) > 0:
        for exclude_field in exclude_fields:
            if exclude_field in dict_to_md5:
                exclude_dict[exclude_field] = dict_to_md5[exclude_field]
                del dict_to_md5[exclude_field]
    sorted_json_str = json.dumps(dict_to_md5, sort_keys=True, ensure_ascii=False)
    return hashlib.md5(sorted_json_str.encode()).hexdigest()



import sqlite3


class SqliteStorage(object):
    def __init__(self, base_dir):
        self.base_dir = base_dir

    def load_persona(self, name):
        return {}

    def load_state(self, name):
        return {}


from isek_config import config
from .file_storage import FileStorage
from .sqlite_storage import SqliteStorage

storage_name = config.get("storage", "enable")
storage = None

if storage_name == "file":
    storage = FileStorage(config.get("storage", "file", "base_dir"))
elif storage_name == "sqlite":
    storage = SqliteStorage(config.get("storage", "sqlite", "base_dir"))



from flask import Flask, request, jsonify, Blueprint
import threading
import time

isek_center_blueprint = Blueprint('isek_center_blueprint', __name__, url_prefix='/isek_center')

nodes = {}

LEASE_DURATION = 30


class CommonResponse(object):
    def __init__(self, data, code, message):
        self.data = data
        self.code = code
        self.message = message

    @classmethod
    def success(cls, data=None, code=200, message='success'):
        return cls(data, code, message).to_dict()

    @classmethod
    def fail(cls, data=None, code=500, message='fail'):
        return cls(data, code, message).to_dict()

    @classmethod
    def error(cls, e):
        return cls(None, e.code, e.message).to_dict()

    def to_dict(self):
        return {
            "code": self.code,
            "message": self.message,
            "data": self.data
        }


@isek_center_blueprint.route('/register', methods=['POST'])
def register():
    data = request.json
    node_id = data.get('node_id')
    host = data.get('host')
    port = data.get('port')
    metadata = data.get('metadata')

    if not node_id or not host or not port:
        return CommonResponse.fail(message="node_id and host/port are required", code=400)

    nodes[node_id] = {
        "node_id": node_id,
        "host": host,
        "port": port,
        "metadata": metadata,
        "expires_at": time.time() + LEASE_DURATION
    }

    return CommonResponse.success()


@isek_center_blueprint.route('/deregister', methods=['POST'])
def deregister():
    data = request.json
    node_id = data.get('node_id')

    if not node_id or node_id not in nodes:
        return CommonResponse.fail(message="Invalid node_id", code=400)

    nodes.pop(node_id)

    return CommonResponse.success()


@isek_center_blueprint.route('/available_nodes', methods=['GET'])
def get_available_nodes():
    current_time = time.time()
    available_nodes = {k: v for k, v in nodes.items() if v['expires_at'] > current_time}
    response = {
        "available_nodes": available_nodes
    }
    return CommonResponse.success(response)


@isek_center_blueprint.route('/renew', methods=['POST'])
def renew():
    data = request.json
    node_id = data.get('node_id')

    if not node_id or node_id not in nodes:
        return CommonResponse.fail(message="Invalid node_id", code=400)

    nodes[node_id]['expires_at'] = time.time() + LEASE_DURATION

    return CommonResponse.success()


def cleanup_expired_nodes():
    while True:
        current_time = time.time()
        expired_nodes = [k for k, v in nodes.items() if v['expires_at'] <= current_time]

        for node_id in expired_nodes:
            nodes.pop(node_id, None)

        time.sleep(5)


def main():
    cleanup_thread = threading.Thread(target=cleanup_expired_nodes, daemon=True)
    cleanup_thread.start()
    app = Flask(__name__)
    app.register_blueprint(isek_center_blueprint)
    app.run(host='0.0.0.0', port=8088, debug=False)

if __name__ == '__main__':
    main()



from isek.node.noderpc import node_pb2, node_pb2_grpc


class IsekNodeServiceServicer(node_pb2_grpc.IsekNodeServiceServicer):

    def __init__(self, isek_node):
        self.isek_node = isek_node

    def send_message(self, request, context):
        # 返回消息
        return node_pb2.CallResponse(reply=f"Hello, {request.message}, I am {self.isek_node.node_id}!")



import ecdsa
from isek.node.etcd_registry import EtcdRegistry


node_a = EtcdRegistry(host="47.236.116.81", port=2379)
node_a.register_node("node_a", "localhost", 8080, {"name": "a"})

node_b = EtcdRegistry(host="47.236.116.81", port=2379)
node_b.register_node("node_b", "localhost", 8082, {"name": "b"})

node_c = EtcdRegistry(host="47.236.116.81", port=2379, parent_node_id="abc")
node_c.register_node("node_c", "localhost", 8082, {"name": "c"})

print(node_b.get_available_nodes())
try:
    node_b.deregister_node("node_a")
except ValueError as e:
    print(f"node_b can not deregister node_a: {e}")
print(node_b.get_available_nodes())
node_a.deregister_node("node_a")
print("node_a deregister node_a")
print(node_b.get_available_nodes())
print(node_c.get_available_nodes())


# public_registry = EtcdRegistry(host="47.236.116.81", port=2379)
# public_registry.register_node("node_1", "localhost", 8080, {"name": "a"})
# public_registry.register_node("node_2", "localhost", 8082, {"name": "b"})
#
# print(public_registry.get_available_nodes())
# public_registry.deregister_node("node_1")
# print(public_registry.get_available_nodes())



from isek.node.etcd_registry import EtcdRegistry
from isek.node.node import Node


class DefaultNode(Node):

    def metadata(self):
        return {}

    def on_message(self, sender, message):
        return f"{self.node_id}"


registry = EtcdRegistry(host="47.236.116.81", port=2379)
node_a = DefaultNode(host="localhost", port=8080, registry=registry)
node_b = DefaultNode(host="localhost", port=8081, registry=registry)

# time.sleep(2)
# node_a.registry.get_available_nodes()





from isek.node.etcd_registry import EtcdRegistry


registry = EtcdRegistry(host="47.236.116.81", port=2379)
nodes = registry.get_available_nodes()
for key, value in nodes.items():
    registry.deregister_node(key)



"""encoding=utf-8"""

from abc import ABC, abstractmethod
from typing import Optional


class AbstractEmbedding(ABC):

    def __init__(self, dim: Optional[int]):
        self.dim = dim

    def embedding_one(self, data: str) -> list[float]:
        return self.embedding([data])[0]

    @abstractmethod
    def embedding(self, datas: list[str]) -> list[list[float]]:
        pass



import json
from typing import Optional, Dict

import etcd3
import base64
from ecdsa.keys import SigningKey, VerifyingKey
from ecdsa.curves import NIST256p

from isek.util.logger import logger
from isek.node.registry import Registry


class EtcdRegistry(Registry):
    def __init__(self,
                 host: Optional[str] = None,
                 port: Optional[int] = None,
                 parent_node_id: Optional[str] = "root",
                 etcd_client: Optional[etcd3.Etcd3Client] = None,
                 ttl: int = 30):

        if host and port and etcd_client:
            logger.warning("Both 'host/port' and 'etcd_client' provided. Using 'etcd_client'.")

        if etcd_client:
            self.etcd_client = etcd_client
        elif host and port:
            self.etcd_client = etcd3.client(host=host, port=port)
        else:
            raise TypeError("Either 'host' and 'port' or 'etcd_client' must be provided.")

        self.parent_node_id = parent_node_id or "root"

        if not self.etcd_client.status():
            raise ConnectionError("Failed to connect to the etcd server.")

        self.sk = SigningKey.generate(curve=NIST256p)

        # 设置租约生存时间（秒）
        self.ttl = ttl
        self.leases = {}

    def register_node(self, node_id: str, host: str, port: int, metadata: Optional[Dict[str, str]] = None):
        """注册节点到注册中心，使用租约保证节点自动过期"""
        vk = self.sk.verifying_key
        vk_bytes = vk.to_string()
        vk_base64 = base64.b64encode(vk_bytes).decode('utf-8')

        node_info = {
            "node_id": node_id,
            "host": host,
            "port": port,
            "public_key": vk_base64,
            "metadata": metadata or {}
        }

        node_info_json = json.dumps(node_info, sort_keys=True).encode("utf-8")
        signature = base64.b64encode(self.sk.sign_deterministic(node_info_json)).decode("utf-8")

        node_entry = {
            "node_info": node_info,
            "signature": signature
        }

        # 创建租约，设置TTL，自动删除失效节点
        lease = self.etcd_client.lease(self.ttl)
        self.leases[node_id] = lease

        key = f"/{self.parent_node_id}/{node_id}"
        self.etcd_client.put(key, json.dumps(node_entry), lease=lease)

        logger.info(f"Node {node_id} registered with info: {node_info}")

    def lease_refresh(self, node_id: str):
        """启动后台线程进行自动续约"""
        lease_refresh_response = None
        try:
            self.__verify_signature(node_id)
            self.leases[node_id].refresh()
            # logger.debug(f"Lease renewed for node: {node_id}, response: {lease_refresh_response}")
        except Exception as e:
            logger.exception(f"Lease renewal failed for node {node_id}, response: {lease_refresh_response}: {e}")

    def get_available_nodes(self) -> Dict[str, dict]:
        """获取当前可用节点列表"""
        nodes = {}
        for value, metadata in self.etcd_client.get_prefix(f"/{self.parent_node_id}/"):
            node_id = metadata.key.decode("utf-8").split(f"/{self.parent_node_id}/")[-1]
            try:
                nodes[node_id] = json.loads(value.decode("utf-8"))['node_info']
            except Exception as e:
                logger.exception(f"Error decoding node {node_id}: {e}")
        return nodes

    def deregister_node(self, node_id: str):
        """从注册中心移除节点并撤销租约"""
        key = f"/{self.parent_node_id}/{node_id}"
        self.__verify_signature(node_id)
        self.etcd_client.delete(key)
        if node_id in self.leases:
            self.leases[node_id].revoke()
            del self.leases[node_id]
        logger.info(f"Node {node_id} deregistered.")

    def __verify_signature(self, node_id):
        key = f"/{self.parent_node_id}/{node_id}"
        node_entry_json = self.etcd_client.get(key)[0]

        if not node_entry_json:
            raise ValueError(f"Node {node_id} not found")

        node_entry = json.loads(node_entry_json.decode("utf-8"))
        node_info = node_entry["node_info"]
        node_base64_signature = node_entry["signature"]

        vk_bytes = base64.b64decode(node_info["public_key"])
        vk = VerifyingKey.from_string(vk_bytes, curve=NIST256p)

        node_info_json = json.dumps(node_info, sort_keys=True).encode("utf-8")
        calculated_signature = self.sk.sign_deterministic(node_info_json)
        calculated_base64_signature = base64.b64encode(calculated_signature).decode("utf-8")

        if calculated_base64_signature != node_base64_signature:
            raise ValueError(f"Signature verification failed for node {node_id}!")

        if not vk.verify(calculated_signature, node_info_json):
            raise ValueError(f"Node {node_id} has an invalid signature!")



"""encoding=utf-8"""

import time
import os
import json
from isek.util.logger import logger
from isek.llm.abstract_model import AbstractModel
from isek.util.tools import function_to_schema, load_json_from_chat_response
from typing import Union, List, Optional, Dict, Callable
from openai import OpenAI





class OpenAIModel(AbstractModel):
    """
    OpenAIModel
    """

    def __init__(
            self,
            model_name: Optional[str] = "gpt-4o-mini",
            api_key: Optional[str] = None,
            base_url: Optional[str] = None
    ):
        super().__init__()
        self.model_name = model_name or "gpt-4o-mini"
        self.client = OpenAI(base_url=base_url, api_key=api_key)

    def generate_json(self, prompt, system_messages=None, retry=3, check_json_def=None):
        for i in range(retry):
            try:
                response = self.create(messages=[{'role': 'user', 'content': prompt}],
                                       systems=system_messages)
                response_content = response.choices[0].message.content
                try:
                    json_result = json.loads(response_content)
                except:
                    json_result = load_json_from_chat_response(response_content)
                if check_json_def:
                    check_json_def(json_result)
                return json_result
            except:
                logger.exception(f"Request model[{self.model_name}] generate_json call fail {i} times")
        raise RuntimeError("call ernie bot error over 3 times")

    def generate_text(self, prompt, system_messages=None, retry=3):
        for i in range(retry):
            try:
                response = self.create(messages=[{'role': 'user', 'content': prompt}], systems=system_messages)
                return response.choices[0].message.content
            except:
                logger.exception(f"Request model[{self.model_name}] generate_text call fail {i} times")
        raise RuntimeError("call ernie bot error over 3 times")

    def create(
            self,
            messages: Union[List[Dict]],
            systems: Optional[List[Dict]],
            tool_schemas: List[Dict] = None
    ):
        try:
            messages = (systems if systems else []) + messages

            logger.debug(f"Request model[{self.model_name}] messages: {messages}")
            start_time = time.time()
            response = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                tools=tool_schemas
            )
            cost_seconds = time.time() - start_time
            logger.debug(f"Request model[{self.model_name}] time taken[{cost_seconds:.2f}s] response[{response}]")
            return response
        except Exception as e:
            logger.exception(f"Request model[{self.model_name}] error.")
            raise e



import os
from isek.agent.distributed_agent import DistributedAgent
from isek.llm.openai_model import OpenAIModel
from isek.node.isek_center_registry import IsekCenterRegistry
from dotenv import load_dotenv


def main():
    load_dotenv()
    model = OpenAIModel(
        model_name=os.environ.get("OPENAI_MODEL_NAME"),
        base_url=os.environ.get("OPENAI_BASE_URL"),
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    
    registry = IsekCenterRegistry()
    Infy_agent = DistributedAgent(model=model)
    Infy_agent.build(daemon=True)
    Infy_agent.run_cli()

if __name__ == "__main__":
    main()
    



import threading
from isek.agent.abstract_agent import AbstractAgent


class SingleAgent(AbstractAgent):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    def build(self, daemon=False):
        if not daemon:
            self.run_cli()
        else:
            termination_thread = threading.Thread(target=self.run_cli, daemon=True)
            termination_thread.start()





import datetime
import uuid
from typing import Dict, List, Any, Optional, Callable, TypeVar, Union
from isek.util.logger import logger
import json

T = TypeVar('T')

class AgentMemory:
    """
    Manages the complete cognitive state of an agent, including working memory,
    long-term knowledge, scheduled tasks, and internal state variables.
    """
    
    def __init__(self):
        """
        Initialize the agent's memory systems.
        
        Args:
        """
        self.logger = logger
        
        # Core memory components
        self.state_variables: Dict[str, Any] = {"is_registered": False, "goal": None}
        self.working_memory: List[str] = []
        self.task_queue: Dict[str, Dict[str, Any]] = {}
        self.knowledge_store: Dict[str, Dict[str, str]] = {}
    
    def log_operation(self, operation_name: str, details: str) -> None:
        """Log a memory operation if a logger is available."""
        if self.logger:
            self.logger.debug(f"Memory: {operation_name} - {details}")
    
    def generate_timestamp(self) -> str:
        """Generate a current timestamp string."""
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    
    def generate_unique_id(self) -> str:
        """Generate a unique ID for memory items."""
        return str(uuid.uuid4())[:8]
    
    # State variable methods
    def set_state_variable(self, variable_name: str, value: Any) -> str:
        """
        Set or update a state variable.
        
        Args:
            variable_name: Name of the state variable to update
            value: New value for the state variable
            
        Returns:
            str: Confirmation message
        """
        self.state_variables[variable_name] = value
        self.log_operation("set_state_variable", f"{variable_name}: {value}")
        return f"State variable '{variable_name}' set to '{value}'"
    
    def get_state_variable(self, variable_name: str, default: T = None) -> Union[Any, T]:
        """
        Get a state variable value.
        
        Args:
            variable_name: Name of the state variable to retrieve
            default: Default value if variable doesn't exist
            
        Returns:
            The variable value or default if not found
        """
        return self.state_variables.get(variable_name, default)
    
    def get_all_state_variables(self) -> Dict[str, Any]:
        """
        Get all state variables.
        
        Returns:
            Dict: All state variables and their values
        """
        self.log_operation("get_all_state_variables", f"count: {len(self.state_variables)}")
        return self.state_variables.copy()
    
    # Working memory methods
    def store_memory_item(self, content: str) -> str:
        """
        Add an item to working memory.
        
        Args:
            content: Content to store in working memory
            
        Returns:
            str: Confirmation message
        """
        self.working_memory.append(content)
        self.log_operation("store_memory_item", f"item: {content[:50]}...")
        return f"Memory item stored"
    
    def get_recent_memory_items(self, count: int = 4) -> List[str]:
        """
        Get the most recent items from working memory.
        
        Args:
            count: Maximum number of items to retrieve
            
        Returns:
            List[str]: The most recent memory items
        """
        if len(self.working_memory) > count:
            return self.working_memory[-count:]
        return self.working_memory.copy()
    
    # Task queue methods
    def create_task(self, task_description: str, priority: int = 1) -> str:
        """
        Create a new task in the task queue.
        
        Args:
            task_description: Description of the task
            priority: Task priority (higher number = higher priority)
            
        Returns:
            str: ID of the created task
        """
        task_id = self.generate_unique_id()
        self.task_queue[task_id] = {
            "description": task_description,
            "created_at": self.generate_timestamp(),
            "completed": False,
            "priority": priority
        }
        self.log_operation("create_task", f"id: {task_id}, task: {task_description}")
        return task_id
    
    def mark_task_completed(self, task_id: str) -> bool:
        """
        Mark a task as completed.
        
        Args:
            task_id: ID of the task to mark as completed
            
        Returns:
            bool: True if successful, False if task not found
        """
        if task_id in self.task_queue:
            self.task_queue[task_id]["completed"] = True
            self.task_queue[task_id]["completed_at"] = self.generate_timestamp()
            self.log_operation("mark_task_completed", f"id: {task_id}")
            return True
        return False
    
    def get_pending_tasks(self) -> Dict[str, Dict[str, Any]]:
        """
        Get all pending (incomplete) tasks.
        
        Returns:
            Dict: Dictionary of pending tasks with their IDs as keys
        """
        pending = {
            task_id: task_info for task_id, task_info in self.task_queue.items() 
            if not task_info.get("completed", False)
        }
        self.log_operation("get_pending_tasks", f"count: {len(pending)}")
        return pending
    
    def get_all_tasks(self) -> Dict[str, Dict[str, Any]]:
        """
        Get all tasks, both pending and completed.
        
        Returns:
            Dict: Dictionary of all tasks with their IDs as keys
        """
        return self.task_queue.copy()
    
    # Knowledge store methods
    def store_knowledge(self, topic: str, content: str) -> str:
        """
        Store a knowledge item.
        
        Args:
            topic: Topic or title for the knowledge item
            content: Content of the knowledge item
            
        Returns:
            str: ID of the stored knowledge item
        """
        knowledge_id = self.generate_unique_id()
        self.knowledge_store[knowledge_id] = {
            "topic": topic,
            "content": content,
            "created_at": self.generate_timestamp()
        }
        self.log_operation("store_knowledge", f"id: {knowledge_id}, topic: {topic}")
        return knowledge_id
    
    def retrieve_knowledge(self, topic: str) -> Optional[str]:
        """
        Retrieve knowledge on a specific topic, generating it if not found.
        
        Args:
            topic: Topic to retrieve knowledge about
            generator_func: Optional function to generate knowledge if not found
            
        Returns:
            str: The knowledge content or None if not found and couldn't be generated
        """
        # Search existing knowledge
        for k_id, item in self.knowledge_store.items():
            if item["topic"] == topic:
                self.log_operation("retrieve_knowledge", f"found - topic: {topic}")
                return item["content"]
        
        self.log_operation("retrieve_knowledge", f"not found - topic: {topic}")
        return None
    
    def get_all_knowledge_topics(self) -> List[str]:
        """
        Get all knowledge topics.
        
        Returns:
            List[str]: List of all knowledge topics
        """
        return [item["topic"] for item in self.knowledge_store.values()]
    
    def state_interaction(self, action, title, value):
        """
        This is state manager. Call this function to update or add new state or milestone.
        """
        if action == "update":
            self.set_state_variable(title, value)
            return f"State updated with {title} as {value}"
        elif action == "insert":
            self.set_state_variable(title, value)
            return f"State insert with {title} as {value}"
        elif action == "fetch":
            return f"Your state is {str(self.get_all_state_variables())}"
        else:
            return "No action is taken"
        
    def knowledge_interaction(self, action, content):
        """
        This is knowledge tool. Call this function to get information.
        """
        if action == "query":
           self.retrieve_knowledge(content)
        elif action == "insert":
            # TODO: topic should be extracted from the content
            self.store_knowledge(content, content)
            return f"Knowledge {content} inserted"
        else:
            return "No action is taken"

        
    def scheduling(self, option, id = None, task=None, finished=False):
        """
        This is scheduling manager. Call this function to add new task to your schedule.
        """
        if option == "add":
            task_id = self.create_task(task)
            return f"Task {task} is added to your schedule, task id is {task_id}"
        
        elif option == "complete":
            self.mark_task_completed(id)
            return f"Task {task} is updated in your schedule, task id is {task_id}"
        
        elif option == "fetch":
            return f"Your schedule is {str(self.get_all_tasks())}"
        else:
            return "No action is taken"


import time
import os
import sys

from isek.agent.single_agent import SingleAgent
from isek.llm.openai_model import OpenAIModel
from isek.agent.persona import Persona
from dotenv import load_dotenv


def combination_calculator(n: str, k: str) -> str:
    # non-recursive implementation
    n = int(n)
    k = int(k)
    if k == 0:
        return "1"
    if k > n:
        return "0"
    result = 1
    for i in range(k):
        result = result * (n - i) // (i + 1)
    return str(result)


def add(a: str, b: str) -> str:
    return str(int(a) + int(b))


def subtract(a: str, b: str) -> str:
    return str(int(a) - int(b))


def multiply(a: str, b: str) -> str:
    return str(int(a) * int(b))


def divide(a: str, b: str) -> str:
    return str(int(a) / int(b))


# model = OpenAIModel(
#     model_name="deepseek-chat",
#     base_url="https://api.deepseek.com",
#     api_key=os.environ.get("DEEPSEEK_API_KEY")
# )


def main():
    load_dotenv()

    model = OpenAIModel(
        model_name=os.environ.get("OPENAI_MODEL_NAME"),
        base_url=os.environ.get("OPENAI_BASE_URL"),
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    
    Eddie_info = {
    "name": "Eddie",
    "bio": "An experienced card game player",
    "lore": "Your mission is to assist user with card game related questions",
    "knowledge": "probability calculation, card game rules, card game strategies",
    "routine": "if you receive tasks that are not related to card game, reject the task"
    }
    
    Eddie = Persona.from_json(Eddie_info)
    
    eddie_agent = SingleAgent(model=model, persona = Eddie)
    eddie_agent.tool_manager.register_tools([combination_calculator, add, subtract, multiply, divide])
    time.sleep(2)
    eddie_agent.run_cli()
    # what is the chance of picking 4 cards and they are 4 Ace in card game

if __name__ == "__main__":
    main()



import os
import time
from isek.isek_config import IsekConfig
from isek.llm import OpenAIModel
from dotenv import load_dotenv


def main():
    load_dotenv()
    model = OpenAIModel(
        model_name=os.environ.get("OPENAI_MODEL_NAME"),
        base_url=os.environ.get("OPENAI_BASE_URL"),
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    
    sylana_config = IsekConfig("examples/sylana/sylana_config.yaml")
    sylana_agent = sylana_config.load_agent()
    sylana_agent.model = model
    sylana_agent.build(daemon=True)

    dobby_config = IsekConfig("examples/dobby/dobby_config.yaml")
    dobby_agent = dobby_config.load_agent()
    sylana_agent.model = model
    dobby_agent.build(daemon=True)

    time.sleep(5)
    hello = f"Hello! My name is {sylana_agent.persona.name}"
    print(f"{sylana_agent.persona.name}: {hello}")

    response = sylana_agent.send_message(dobby_agent.node_id, hello)
    print(f"{dobby_agent.persona.name}: {response}")


if __name__ == '__main__':
    main()



from isek.agent.abstract_agent import AbstractAgent
from isek.node.node import Node
from isek.util.logger import logger
import threading


class DistributedAgent(AbstractAgent, Node):

    def __init__(
            self,
            **kwargs
    ):
        # 调用 AbstractAgent 的构造方法
        AbstractAgent.__init__(self, **kwargs)
        # 生成 intro
        self.intro = self.persona.bio
        # self.intro_vector = self.embedding.embedding_one(self.intro)
        # 调用 Node 的构造方法
        Node.__init__(self, **kwargs)

    def build(self, daemon=False):
        if not daemon:
            self.build_server()
        else:
            termination_thread = threading.Thread(target=self.build_server, daemon=True)
            termination_thread.start()

    def build_node_id(self) -> str:
        return self.persona.name

    def metadata(self):
        return {
            "name": self.persona.name,
            "intro": self.intro
        }

    def on_message(self, sender, message):
        logger.info(f"[{self.persona.name}] received message from {sender}: {message}")

        return self.run(message)

    def search_partners(self, query: str) -> str:
        """
        in case of lacking knowledge to answer the query, search for partners based on the query,
        Args:
            query: str
        Returns: partner name and node_id
        """
        logger.info(f"[{self.persona.name}] Searching partners with query: {query}")
        nodes = str(self.get_nodes_by_vector(query))
        # 前期node少，用语意理解来匹配 后期vector来匹配
        
        matching_node_template = f"""
            I am looking for partners to help me with a query, here are the nodes I found:
            {nodes}
            this is the query:
            {query}

            please return the node name and node_id of only one node that fit best with the query based on he intro of that node.
            reason always less than 10 words.

            return in the format of:
            name: XXX, node_id: XXX, Reason: XXX

        """
        result = self.model.generate_text(prompt=matching_node_template)
        return result


from abc import ABC, abstractmethod
from typing import Optional, Dict


class Registry(ABC):

    @abstractmethod
    def register_node(self, node_id: str, host: str, port: int, metadata: Optional[Dict[str, str]] = None):
        """注册节点到注册中心"""
        pass

    @abstractmethod
    def get_available_nodes(self) -> dict:
        """获取当前可用节点列表"""
        pass

    @abstractmethod
    def deregister_node(self, node_id: str):
        """从注册中心移除节点"""
        pass

    @abstractmethod
    def lease_refresh(self, node_id: str):
        pass



import json
from typing import Optional, Dict

import requests

from isek.util.logger import logger
from isek.node.registry import Registry


class IsekCenterRegistry(Registry):
    def __init__(self,
                 host: Optional[str] = "localhost",
                 port: Optional[int] = 8088,
                 ):
        self.center_address = f"http://{host}:{port}"
        self.node_info = {}

    def register_node(self, node_id: str, host: str, port: int, metadata: Optional[Dict[str, str]] = None):
        """注册节点到注册中心，使用租约保证节点自动过期"""

        register_url = f"{self.center_address}/isek_center/register"
        self.node_info = {
            "node_id": node_id,
            "host": host,
            "port": port,
            "metadata": metadata or {}
        }

        response = requests.post(url=register_url, json=self.node_info)
        response_json = json.loads(response.content)
        if response_json['code'] != 200:
            raise RuntimeError(f'Register isek center error {response_json}')
        logger.debug(f"Node {node_id} registered, response info: {response.content}")

    def lease_refresh(self, node_id: str):
        lease_refresh_url = f"{self.center_address}/isek_center/renew"
        node_info = {
            "node_id": node_id,
        }

        response = requests.post(url=lease_refresh_url, json=node_info)
        response_json = json.loads(response.content)
        if response_json['code'] != 200:
            raise RuntimeError(f'deregister from isek center error {response_json}')
        # logger.debug(f"Node {node_id} lease refresh.")


    def get_available_nodes(self) -> Dict[str, dict]:
        """获取当前可用节点列表"""
        register_url = f"{self.center_address}/isek_center/available_nodes"
        response = requests.get(url=register_url)
        response_json = json.loads(response.content)
        if response_json['code'] != 200:
            raise RuntimeError(f'Get available nodes from isek center error {response_json}')
        return response_json['data']['available_nodes']

    def deregister_node(self, node_id: str):
        """从注册中心移除节点并撤销租约"""
        register_url = f"{self.center_address}/isek_center/deregister"
        node_info = {
            "node_id": node_id,
        }

        response = requests.post(url=register_url, json=node_info)
        response_json = json.loads(response.content)
        if response_json['code'] != 200:
            raise RuntimeError(f'deregister from isek center error {response_json}')
        logger.info(f"Node {node_id} deregistered.")



from isek.llm.openai_model import OpenAIModel
import os
from dotenv import load_dotenv
from isek.util.logger import LoggerManager

def main():
    load_dotenv()
    LoggerManager.init(debug=True)
    model = OpenAIModel(
        model_name=os.environ.get("OPENAI_MODEL_NAME"),
        base_url=os.environ.get("OPENAI_BASE_URL"),
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    text = "Do androids dream of electric sheep?"
    response = model.generate_text(text)
    print(f"Bot: {response}")

if __name__ == "__main__": 
    main()



from abc import ABC, abstractmethod
from isek.llm.openai_model import OpenAIModel
from isek.embedding.openai_embedding import OpenAIEmbedding
from isek.llm.abstract_model import AbstractModel
from isek.embedding.abstract_embedding import AbstractEmbedding
from typing import Optional, List, Callable
from isek.agent.persona import Persona
from isek.agent.memory import AgentMemory
from isek.agent.toolbox import ToolBox
import time
# add logging
from isek.util.logger import LoggerManager, logger


class AbstractAgent(ABC):
    def __init__(
            self,
            persona: Persona = None,
            model: Optional[AbstractModel] = None,
            tools: List[Callable] = None,
            heartbeat: bool = False,
            **kwargs
    ) -> None:
        self.persona = persona or Persona.default()
        self.model = model or OpenAIModel()
        self.memory_manager = AgentMemory()
        self.tool_manager = ToolBox(persona=self.persona)
        # Register action tools
        if tools:
            self.tool_manager.register_tools(tools)

    @abstractmethod
    def build(self, daemon=False):
        pass

    def run(self, input : str = None) -> None:
        """
        Args:
            input:
                user instructions or environment signals, such as text, image, file, web page, etc.
        Returns:
            None
        """
        logger.info(f"[{self.persona.name}] ++++++++++Cycle Started++++++++++")
        response = self.response(input)
        logger.info(f"[{self.persona.name}][Response]: {response}")
        logger.info(f"[{self.persona.name}] ----------Cycle Ended------------")
        
        return response
    
    def hearbeat(self):
        """
        This is a heartbeat function. Call this function when you want to keep the agent alive.
        """
        self.run()
        
    def run_cli(self):
        while True:
            time.sleep(0.2)
            text = input("Enter your command: ")
            if text == "exit":
                break
            self.run(text)
            
    def response(self, input: str) -> str:
        """
        Response to the input
        Args:
            input: user instructions or environment signals, such as text, image, file, web page, etc.
        Returns:
            str: response to the input
        """
        if input is not None and input != "":
            logger.info(f"[{self.persona.name}][Trigger: Input]: {input}")
        else:
            logger.info(f"[{self.persona.name}][Trigger: Heartbeat]")
        self.memory_manager.store_memory_item("User:" + input)
        # Build template for action phase
        template = self._build_templates()
        messages = []
        
        # Setup available tools for action phase
        tools = self.tool_manager.get_tool_names()
        tool_schemas = self.tool_manager.get_tool_schemas()
        
        # Main action loop
        while True:
            # Get AI completion with tool calling
            response = self.model.create(
                messages=[{"role": "system", "content": template}] + messages,
                systems=[],
                tool_schemas=tool_schemas or None,
            ).choices[0].message
            messages.append(response)
            
            # Process text response
            if response.content:
                self.memory_manager.store_memory_item("Agent:" + response.content)
                
            # Check if we're done with tool calls
            if not response.tool_calls:
                return response.content

            # Handle tool calls
            for tool_call in response.tool_calls:
                result = self.tool_manager.execute_tool_call(tool_call=tool_call)
                result_message = {
                    "role": "tool",
                    "tool_call_id": tool_call.id,
                    "content": result,
                }
                messages.append(result_message)
                
    def _build_templates(self):
        """Build templates for the action phase."""
        # Get recent memory items
        recent_memory = self.memory_manager.get_recent_memory_items()
        # state = self.memory_manager.get_all_state_variables()
        # tasks = self.memory_manager.get_pending_tasks()
       
        # Return all templates as a dictionary
        routine=""
        if "routine" in self.persona.__dict__:
            routine = self.persona.routine
        
        templates = {
            "persona": f'''       
                "You Name is:
                "{self.persona.name}"
                "You are a {self.persona.bio}"
                "Your mission is to {self.persona.lore}"
                "You have the following knowledge: {self.persona.knowledge}"
            ''',
            "system": f'''
                "You have the freedom to reason and act to help you achieve your goal."
                "try your best to use tools to improve responses quality."
                "please be concise and clear in your responses."
                "in most cases, you should only provide one or two sentence."
                "in rare cases, you may provide a longer response." 
            ''',
            "routine": f'''
            "{routine}"
            ''',
            "tool": f'''
                "You have the following tools available: {self.tool_manager.get_tool_names()}"
            ''',
            
            "closing": f'''
                response base on the following conversation: 
                {recent_memory}
            '''
        }
        # add all the templates together
        template = ""
        for key in templates:
            template += templates[key]
        return template
        
    def decompose_task(self, command):
        """
        This is message decomposer. Call this function when you want to decompose a task to a list of subtasks.
        """
        task_decomposer_template = f"""
        As a Task Decomposer, your objective is to divide the given task into high level subtasks.
        You have been provided with the following objective:
        {command}

        Please format the subtasks as a python list with 3 or less tasks, each tasks contains less than 10 words, as demonstrated below, output nothing else:
        ["Subtask 1", "Subtask 2", "Subtask 3"]
        
        Each subtask should be concise, concrete, and achievable.
        Ensure that the task plan is created without asking any questions.
        Be specific and clear.
        """
        response = self.model.create(
                messages=[{"role": "system", "content": task_decomposer_template}],
                systems=[],
                tool_schemas= None,
            ).choices[0].message

        # parse the response to get the subtasks

        subtasks = eval(response.content)

        # for subtask in subtasks:
        #   self.memory_manager.create_task(subtask)
        logger.info(f"[{self.persona.name}] Subtasks: {subtasks}")
        return str(subtasks) 



import json
import threading
from abc import ABC, abstractmethod
from concurrent import futures
from typing import Dict

import faiss
import copy
import grpc
import numpy as np

from isek.constant.exceptions import NodeUnavailableError
from isek.node.noderpc import node_pb2, node_pb2_grpc
from isek.node.registry import Registry
from isek.util.logger import logger
from isek.node.node_index import NodeIndex
from isek.embedding.abstract_embedding import AbstractEmbedding
from isek.node.isek_center_registry import IsekCenterRegistry


class Node(node_pb2_grpc.IsekNodeServiceServicer, ABC):
    def __init__(self,
                 host: str = "localhost",
                 port: int = 8080,
                 registry: Registry = IsekCenterRegistry(),
                 embedding: AbstractEmbedding = None,
                 **kwargs
                 ):
        if not host or not port or not registry:
            raise ValueError("Node")
        self.node_id = self.build_node_id()
        self.host = host
        self.port = port
        self.registry = registry
        self.all_nodes = {}
        self.node_index = None
        if embedding:
            self.node_index = NodeIndex(embedding)
        self.node_list = None
        # self.__build_server()

    @abstractmethod
    def build_node_id(self) -> str:
        pass

    @abstractmethod
    def metadata(self) -> Dict:
        pass

    @abstractmethod
    def on_message(self, sender, message) -> str:
        pass

    def build_server(self):
        self.registry.register_node(node_id=self.node_id, host=self.host, port=self.port, metadata=self.metadata())
        self.__bootstrap_heartbeat()
        self.__bootstrap_grpc_server()

    def __bootstrap_heartbeat(self):
        self.registry.lease_refresh(self.node_id)
        self.__refresh_nodes()
        # logger.debug(f"{self.node_id} get all available nodes: {self.all_nodes}")
        timer = threading.Timer(5, self.__bootstrap_heartbeat)
        timer.daemon = True
        timer.start()

    def __refresh_nodes(self):
        all_nodes = self.registry.get_available_nodes()
        # todo self.node_index.compare_and_build(all_nodes)
        # is_nodes_changed = False
        # for node_id, node_info in all_nodes.items():
        #     if node_id not in self.all_nodes:
        #         is_nodes_changed = True
        #         break
        # if is_nodes_changed or len(all_nodes) != len(self.all_nodes):
        #     node_ids, vectors = zip(*[(node_id, n['metadata']['intro_vector']) for node_id, n in all_nodes.items()])
        #     vector_dim = len(vectors[0])
        #     vectors = np.array(vectors, dtype='float32')
        #     self.node_index = faiss.IndexFlatL2(vector_dim)
        #     self.node_index.add(vectors)
        #     self.node_list = node_ids
        #     logger.debug("Node index rebuild finished.")
        self.all_nodes = all_nodes

    def __bootstrap_grpc_server(self):
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        node_pb2_grpc.add_IsekNodeServiceServicer_to_server(self, server)

        # 监听端口
        server.add_insecure_port(f'[::]:{self.port}')
        server.start()
        logger.info(f"[{self.node_id}] Node started on port {self.port}...")
        server.wait_for_termination()
        # def wait_for_termination():
        #
        #
        # termination_thread = threading.Thread(target=wait_for_termination)
        # termination_thread.start()

    def send_message(self, receiver_node_id, message):
        """
        send message to another node by providing receiver_node_id= agent_name and message = message
        """
        logger.info(f"[{self.node_id}] send msg to [{receiver_node_id}]: {message}")
        receiver_node = self.all_nodes.get(receiver_node_id, None)
        if not receiver_node:
            raise NodeUnavailableError(receiver_node)
        # 连接到 gRPC 服务
        channel = grpc.insecure_channel(f"{receiver_node['host']}:{receiver_node['port']}")
        stub = node_pb2_grpc.IsekNodeServiceStub(channel)

        # 创建请求消息
        request = node_pb2.CallRequest(sender=self.node_id, receiver=receiver_node_id, message=message)

        # 调用远程服务方法
        response = stub.call(request)
        # log the response
        logger.info(f"[{self.node_id}] receive message from [{receiver_node_id}]: {response.reply}")
        return f"{response.reply}"

    def get_nodes_by_vector(self, query, limit=20):
        return self.all_nodes.values()
        # todo
        # if len(self.all_nodes) > limit and self.node_index is not None:
        #     node_ids = self.node_index.search(query, limit=limit)
        #     results = [self.all_nodes[node_id] for node_id in node_ids]
        # else:
        #     results = self.all_nodes.values()
        # return results

    def call(self, request, context):
        # 返回消息
        return node_pb2.CallResponse(reply=self.on_message(request.sender, request.message))



import os
import shutil

def remove_directory(path):
    """Remove a directory if it exists."""
    if os.path.exists(path):
        print(f"Removing directory: {path}")
        shutil.rmtree(path, ignore_errors=True)

def remove_file(path):
    """Remove a file if it exists."""
    if os.path.exists(path):
        print(f"Removing file: {path}")
        os.remove(path)

def main(root='.'):
    unwanted_dirs = []

    # Walk the directory tree
    for dirpath, dirnames, filenames in os.walk(root):
        # Identify __pycache__ directories and directories ending with .egg-info
        for dirname in dirnames:
            full_path = os.path.join(dirpath, dirname)
            if dirname == '__pycache__' or dirname.endswith('.egg-info'):
                unwanted_dirs.append(full_path)

        # Remove any .pyc files
        for filename in filenames:
            if filename.endswith('.pyc'):
                file_path = os.path.join(dirpath, filename)
                remove_file(file_path)

    # Additionally, target top-level directories often created during builds
    for extra in ['build', 'dist', 'interfaces', 'reports','contracts']:
        extra_path = os.path.join(root, extra)
        unwanted_dirs.append(extra_path)

    # Remove the collected directories
    for directory in unwanted_dirs:
        remove_directory(directory)

if __name__ == '__main__':
    main()


from .etcd_registry import EtcdRegistry
from .isek_center_registry import IsekCenterRegistry

__all__ = [
    "EtcdRegistry",
    "IsekCenterRegistry",
    "registries"
]


registries = {
    "etcd": EtcdRegistry,
    "isek_center": IsekCenterRegistry
}



from loguru import logger
import sys
from threading import Lock


class LoggerManager:
    _instance = None
    _lock = Lock()

    def __new__(cls, level="INFO", log_format="{message}"):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
            cls._instance.__init_logger(level, log_format)
        return cls._instance

    def __init_logger(self, level, log_format):
        logger.remove()  # 移除默认的日志处理
        if log_format:
            logger.add(
                sys.stdout,
                level=level,
                colorize=True,
                format=log_format,
                enqueue=True  # 线程安全
            )
        else:
            logger.add(
                sys.stdout,
                level=level,
                colorize=True,
                enqueue=True  # 线程安全
            )

    @classmethod
    def init(cls, debug=False):
        """初始化 logger, 支持动态配置"""
        level = "DEBUG" if debug else "INFO"
        log_format = None if debug else "{message}"
        cls(level=level, log_format=log_format)  # 创建单例

    @staticmethod
    def get_logger():
        return logger


# 默认初始化为 INFO，防止未初始化时无法使用
LoggerManager.init(debug=False)

# 直接导出 logger，供其他模块使用
logger = LoggerManager.get_logger()




class NodeUnavailableError(Exception):
    def __init__(self, node_name, message="Node is unavailable"):
        self.node_name = node_name
        self.message = f"Node '{node_name}' is unavailable: {message}"
        super().__init__(self.message)




import json
import inspect
from isek.agent.persona import Persona
from isek.util.logger import logger
from typing import List, Dict, Callable, Any, Optional, Union

class ToolBox:
    def __init__(self, persona: Persona = None) -> None:
        """
        Initialize a tool manager for an agent.
        
        Args:
        """
        self.logger = logger
        self.persona = persona
        # Tool containers
        self.all_tools: Dict[str, Callable] = {}
        
        # Tool metadata
        self.tool_descriptions: Dict[str, str] = {}
        self.tool_schemas: Dict[str, Dict] = {}
    
    def _log(self, message: str) -> None:
        """Log a message if logger is available."""
        if self.logger:
            self.logger.info(message)
    
    def register_tool(self, func: Callable) -> None:
        """
        Register a new tool function.
        
        Args:
            func: The function to register as a tool
            category: Tool category ("action" or "reasoning")
        """
        name = func.__name__
        
        # Store the function in appropriate containers
        self.all_tools[name] = func
        
        
        # Generate and store the schema
        self.tool_schemas[name] = self._function_to_schema(func)
        
        # Store metadata
        self.tool_descriptions[name] = (func.__doc__ or "").strip()
        
        self._log(f"[{self.persona.name}] Tool: added - {name} ")
    
    def register_tools(self, tools: List[Callable]) -> None:
        """Register multiple tools at once."""
        for tool in tools:
            self.register_tool(tool)
    
    def get_tool(self, name: str) -> Optional[Callable]:
        """Get a tool function by name."""
        return self.all_tools.get(name)
    
    def get_tool_names(self) -> List[str]:
        """Get a list of tool names, optionally filtered by category."""
        return list(self.all_tools.keys())
    
    def get_tool_schemas(self, category: Optional[str] = None) -> List[Dict]:
        """Get tool schemas, optionally filtered by category."""
        return [self.tool_schemas[name] for name in self.all_tools]
    
    def execute_tool_call(self, tool_call, **extra_kwargs) -> str:
        """
        Execute a tool call based on LLM response.
        
        Args:
            tool_call: Tool call object from LLM response
            extra_kwargs: Additional kwargs to pass to the tool function
        
        Returns:
            str: Result of the tool execution
        """
        name = tool_call.function.name
        
        if name not in self.all_tools:
            error_msg = f"Tool {name} not found"
            self._log(f"Tool: error - {error_msg}")
            return error_msg
        
        try:
            func = self.all_tools[name]
            args = json.loads(tool_call.function.arguments)
            
            # Merge any extra kwargs
            for key, value in extra_kwargs.items():
                args[key] = value
                
            self._log(f"[{self.persona.name}][Tool] executing {name}({args})")
            result = func(**args)
            return result
        except Exception as e:
            error_msg = f"Error executing {name}: {str(e)}"
            self._log(f"[{self.persona.name}][Tool] error - {error_msg}")
            return error_msg
    
    def _function_to_schema(self, func: Callable) -> Dict:
        """
        Convert a function to a tool schema for LLM API.
        
        Args:
            func: The function to convert
        
        Returns:
            dict: Schema for the function
        """
        type_map = {
            str: "string",
            int: "integer",
            float: "number",
            bool: "boolean",
            list: "array",
            dict: "object",
            type(None): "null",
        }

        try:
            signature = inspect.signature(func)
        except ValueError as e:
            raise ValueError(
                f"Failed to get signature for function {func.__name__}: {str(e)}"
            )

        parameters = {}
        for param in signature.parameters.values():
            try:
                param_type = type_map.get(param.annotation, "string")
            except KeyError as e:
                raise KeyError(
                    f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
                )
            parameters[param.name] = {"type": param_type}

        required = [
            param.name
            for param in signature.parameters.values()
            if param.default == inspect._empty
        ]

        return {
            "type": "function",
            "function": {
                "name": func.__name__,
                "description": (func.__doc__ or "").strip(),
                "parameters": {
                    "type": "object",
                    "properties": parameters,
                    "required": required,
                },
            },
        }



import yaml
import etcd3
from isek.util.logger import logger, LoggerManager
from isek.agent.persona import Persona
from isek.agent.single_agent import SingleAgent
from isek.agent.distributed_agent import DistributedAgent
from isek.llm.openai_model import OpenAIModel
from isek.embedding.openai_embedding import OpenAIEmbedding
from isek.node import EtcdRegistry, IsekCenterRegistry
from isek.llm import llms
from isek.embedding import embeddings
import etcd3


class IsekConfig:
    def __init__(self, yaml_path):
        try:
            with open(yaml_path, "r") as f:
                self.config = yaml.safe_load(f)
            logger.info(f"Config loaded successfully from {yaml_path}")
        except Exception as e:
            logger.exception(f"Error loading IsekConfig: {e}")
            raise e

    def get(self, *keys):
        value = self.config
        for key in keys:
            value = value.get(key, None)
            if value is None:
                return None
        return value

    def get_sub_config(self, key):
        return self.config.get(key)

    def load_agent(self):
        is_dist = self.get("distributed")
        agent_config = self.get_sub_config("agent")
        persona = Persona.load(agent_config.get("persona_path"))
        debug = agent_config.get("debug")
        LoggerManager.init(debug)
        llm = self.load_llm()
        if not is_dist:
            return SingleAgent(persona=persona, model=llm)

        host = self.get("distributed.server", "host")
        port = self.get("distributed.server", "port")
        registry = self.load_registry()
        embedding = self.load_embedding()
        return DistributedAgent(
            host=host, port=port, registry=registry,
            persona=persona, model=llm, embedding=embedding
        )

    def load_registry(self):
        registry_mode = self.get("registry")
        if registry_mode == "etcd":
            return self.load_etcd_registry()
        else:
            return self.load_isek_center_registry()

    def load_etcd_registry(self):
        etcd_client = etcd3.Etcd3Client(**self.get_sub_config("registry.etcd"))
        return EtcdRegistry(etcd_client=etcd_client)

    def load_isek_center_registry(self):
        return IsekCenterRegistry(**self.get_sub_config("registry.isek_center"))

    def load_llm(self):
        llm_mode = self.get("llm")
        if llm_mode is None:
            return None
        return llms.get(llm_mode)(**self.get_sub_config(f"llm.{llm_mode}"))

    def load_embedding(self):
        embedding_mode = self.get("embedding")
        if embedding_mode is None:
            return None
        return embeddings.get(embedding_mode)(**self.get_sub_config(f"embedding.{embedding_mode}"))



import time
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../isek")))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../examples")))

from agent.distributed_agent import DistributedAgent
from node.etcd_registry import EtcdRegistry
from node.isek_center_registry import IsekCenterRegistry
from agent.persona import Persona

os.environ['OPENAI_API_KEY'] = ""


def main():
    registry = EtcdRegistry(host="47.236.116.81", port=2379)
    # registry = IsekCenterRegistry()

    Mani_info = {
        "name": "Mani",
        "bio": "An experienced manager",
        "lore": "Your mission is to manage the team",
        "knowledge": "recruit, assign tasks, manage tasks, manage projects"
    }

    Doki_info = {
        "name": "Doki",
        "bio": "An experienced doctor",
        "lore": "Your mission is to treat patients",
        "knowledge": "diagnose, treat, prescribe, surgery"
    }
    Brook_info = {
        "name": "Brook",
        "bio": "An experienced bookkeeper",
        "lore": "Your mission is to manage the accounts of the company",
        "knowledge": "bookkeeping, accounting, finance, tax"
    }

    Mani = Persona.from_json(Mani_info)
    Doki = Persona.from_json(Doki_info)
    Brook = Persona.from_json(Brook_info)

    Mani_agent = DistributedAgent(persona=Mani, host="localhost", port=8080, registry=registry)
    Mani_agent.tool_manager.register_tools([
        Mani_agent.search_partners,
        Mani_agent.send_message,
        Mani_agent.decompose_task,
    ])

    Doki_agent = DistributedAgent(persona=Doki, host="localhost", port=8081, registry=registry)
    Brook_agent = DistributedAgent(persona=Brook, host="localhost", port=8082, registry=registry)

    # dobby = Persona.load("examples/dobby.character.json")
    # dobby_agent = DistributedAgent(persona=dobby, host="localhost", port=8081, registry=registry)
    time.sleep(5)

    print(Mani_agent.search_partners("An experienced doctor"))
    print(Mani_agent.search_partners("An experienced bookkeeper"))
    print(Mani_agent.search_partners("I'm sick."))


main()


"""encoding=utf-8"""

import os
from typing import Optional

from openai import OpenAI

from isek.embedding.abstract_embedding import AbstractEmbedding
from isek.util.logger import logger
from isek.util.tools import split_list


class OpenAIEmbedding(AbstractEmbedding):
    """
    OpenAIEmbedding
    """

    def __init__(
            self,
            dim: Optional[int],
            model_name: Optional[str] = "text-embedding-3-small",
            api_key: Optional[str] = None,
            base_url: Optional[str] = None
    ):
        super().__init__(dim)
        self.model_name = model_name
        self.client = OpenAI(base_url=base_url, api_key=api_key)

    def embedding(self, datas: list[str]) -> list[list[float]]:
        data_array = split_list(datas, 16)
        all_results = []
        for sub_datas in data_array:
            result = self.client.embeddings.create(input=sub_datas, model=self.model_name)
            all_results.extend([r.embedding for r in result.data])
        return all_results



from isek.util.logger import logger
import json


class Persona:
    def __init__(self, name, bio, lore, knowledge, routine):
        self.name = name
        self.bio = bio
        self.lore = lore
        self.knowledge = knowledge
        self.routine = routine
        self.agencies = []
        logger.info(f'Created Persona: {self.name}')
    
    def __str__(self):
        return self.name

    @classmethod
    def default(cls):
        return cls.from_json({
            "name": "Helper",
            "bio": "You are good at handling various things",
            "lore": "Your mission is to provide users with various assistance Be sure to reject all requests involving politics, religion, and pornography, and do not provide anything that violates the law or morality.",
            "knowledge": "Provide solutions and implementation steps",
            "routine": "You should first understand the user's request, then provide a solution, and finally ask the user if the solution is satisfactory."
        })
    
    @classmethod
    def load(cls, path):
        try:
            with open(path, 'r') as file:
                data = json.load(file)
                return cls.from_json(data)
        except FileNotFoundError:
            logger.exception(f'File not found: {path}')
            raise

    @classmethod
    def from_json(cls, data):
        try:
            return cls(**data)
        except KeyError:
            logger.exception(f'Invalid data: {data}')
            raise



import faiss
import numpy as np
from isek.util.tools import dict_md5
from isek.util.logger import logger
from isek.embedding.abstract_embedding import AbstractEmbedding


class NodeIndex(object):
    def __init__(self, embedding: AbstractEmbedding):
        self.node_info_dict = {}
        self.node_ids = []
        self.node_index = None
        self.embedding = embedding
        self.dim = self.embedding.dim

    def compare_and_build(self, all_nodes):
        has_vector_nodes = {}
        not_has_vector_nodes = {}
        not_has_vector_intros = []
        for node_id, node_info in all_nodes.items():
            node_md5 = dict_md5(node_info)
            if (node_id not in self.node_info_dict
                    or self.node_info_dict[node_id]['md5'] != node_md5):
                not_has_vector_nodes[node_id] = {
                    "md5": node_md5
                }
                not_has_vector_intros.append(node_info["metadata"]["intro"])
            else:
                has_vector_nodes[node_id] = self.node_info_dict[node_id]

        vectors = self.embedding.embedding(not_has_vector_intros)
        for node_id, vector in zip(not_has_vector_nodes.keys(), vectors):
            not_has_vector_nodes[node_id]['vector'] = vector

        self.node_info_dict = has_vector_nodes | not_has_vector_nodes
        if len(not_has_vector_nodes) > 0:
            self.rebuild()

    def rebuild(self):
        node_ids, vectors = zip(*[(node_id, n['vector']) for node_id, n in self.node_info_dict.items()])
        vectors = np.array(vectors, dtype='float32')
        new_index = faiss.IndexFlatL2(self.dim)
        new_index.add(vectors)
        self.node_index = new_index
        self.node_ids = node_ids
        logger.debug("Node index rebuild finished.")

    def search(self, query, limit=20):
        vector = self.embedding.embedding_one(query)
        results = []
        if len(self.node_ids) > limit:
            query_vector = np.array([vector], dtype='float32')
            distances, indices = self.node_index.search(query_vector, limit)
            for distance, index in zip(distances[0], indices[0]):
                real_distance = 1 - distance / 2.0
                node_id = self.node_ids[index]
                logger.debug(f"search result index[{index}] distance[{real_distance}] node_id[{node_id}]")
                results.append(node_id)
        else:
            results = self.node_ids
        return results